#include <stdio.h>
#include <string.h>

#include "object.h"
#include "util.h"
#include "Attribute.h"
#include "LandscapeObject.h"
#include "LandscapeObject3D.h"
#include "RelationInstance.h"
#include "EntityInstance.h"
#include "EntityCache.h"
#include "TAInput.h"
#include "TAStreamTokenizer.h"
#include "LandscapeTokenStream.h"

#ifdef WIN32
#define stricmp _stricmp
#else
#define stricmp strcasecmp
#endif

void 
CLandscapeTokenStream::skipRecord(void) 
{
	int ttype;

	do {
		ttype = nextToken();
	} while (ttype != TT_EOF && ttype != '}');
}

void 
CLandscapeTokenStream::parseAttributes(CLandscapeObject *targetP) 
{
	int			ttype;
	const char	*attributeIdP;
	const char	*valueP;

	for (;;) {
		ttype = nextToken();

		switch (ttype) {
		case '}':
			return;
		case TT_WORD:
			break;
		default:
			fprintf(stderr, "*** Error (%d): Expected attribute id:. Found '%s'\n", m_lineno, m_svalP);
			return;
		}
		attributeIdP = m_svalP;

		ttype = nextToken();
		if (ttype != '=') {
			// Attribute declaration with no value 
			pushBack(ttype);
			valueP = 0; 
		} else {
			valueP = nextAVI();
		}
		targetP->putAttribute(attributeIdP, valueP);
	}
}

// --------------

// Public methods

// --------------

CLandscapeTokenStream::CLandscapeTokenStream(CTAInput *taInputP)
{
	m_taInputP     = taInputP;
	m_entityCacheP = taInputP->m_entityCacheP;
	m_startLineno  = -1;
	m_token1P      = 0;
	m_token2P      = 0;
	m_ttype3       = 0;
	m_verbP        = 0;
	m_objectP      = 0;
	m_subjectP     = 0;
	m_relations    = 0;
}

int 
CLandscapeTokenStream::getStartLineno(void) const
{
	return m_startLineno;
}

int 
CLandscapeTokenStream::nextSection(void)
{
	// Called when a section id is expected
	// Returns section id if new section found, or 
	// EOF when at stream end.

	int ttype;

	const char *graphP, *typeP;

	if (m_token1P) {
		if (m_ttype3 == ':') {
			graphP    = m_token1P; 
			typeP     = m_token2P;
			m_token1P = 0;
			m_token2P = 0;
		} else {
			fprintf(stderr, "*** Error (%d): Expected section header ':'. Found '%s'\n", m_lineno, m_svalP);
			error_exit(1);
		}
	} else { 
		ttype = nextToken();
		switch (ttype) {
		case TT_EOF:
			return EOFTOKEN;
		case TT_WORD:
			graphP = m_svalP;

			if (nextToken() != TT_WORD) {
				fprintf(stderr, "*** Error (%d): Expecting section type id. Found '%s'\n", m_lineno, m_svalP);
				error_exit(1);
			}
			typeP = m_svalP; 
			if (nextToken() != ':') {
				fprintf(stderr, "*** Error (%d): Expecting ':'. Found '%s'\n", m_lineno, m_svalP);
				error_exit(1);
			}
			break;
		default:
			fprintf(stderr, "*** Error (%d): Expecting graph id. Found '%s'\n", m_lineno, m_svalP);
			error_exit(1);
	}	} 

	if (!stricmp(graphP, "SCHEME")) {
		if (!stricmp(typeP, "TUPLE")) {
			return SCHEME_TUPLE;
		}
		if (!stricmp(typeP, "ATTRIBUTE")) {
			return SCHEME_ATTRIBUTE;
		}
		if (!stricmp(typeP, "END")) {
			return EOFTOKEN;
		}
		fprintf(stderr, "*** Error (%d): Bad section type. Found '%s'\n", m_lineno, m_svalP);
		error_exit(1);
	}

	if (!stricmp(graphP, "FACT")) {
		if (!stricmp(typeP, "TUPLE")) {
			return FACT_TUPLE;
		}
		if (!stricmp(typeP, "ATTRIBUTE")) {
			return FACT_ATTRIBUTE;
		}
		fprintf(stderr, "*** Error (%d): Bad section type. Found '%s'\n", m_lineno, m_svalP);
		error_exit(1);
	}

	fprintf(stderr, "*** Error (%d): Bad section type id. Found '%s'\n", m_lineno, m_svalP);
	error_exit(1);
	return(0);
}

int
CLandscapeTokenStream::nextSchemaTriple(void) 
{
	m_startLineno = lineno();
		
	switch (nextToken()) {
	case TT_EOF:
		// End of section 
		return false;
	case TT_WORD: 
		break;
	default:
		fprintf(stderr, "*** Error (%d): Expecting word verb token. Found '%s'\n", m_lineno, m_svalP);
		error_exit(1);
	}

	m_relations = 0;
	m_verbP     = m_svalP;
	switch (nextToken()) {
	case TT_WORD:
		m_objectP = m_svalP;
		break;
	case '(':
		m_relations |= 1;
		if (nextToken() == TT_WORD) {
			m_objectP = m_svalP;
			if (nextToken() == ')') {
				break;
		}	}
	default:
		fprintf(stderr, "*** Error (%d): Expecting word object token. Found '%s'\n", m_lineno, m_svalP);
		error_exit(1);
	}

	switch (nextToken()) {
	case ':':
		m_token1P = m_verbP;
		m_token2P = m_objectP; 
		m_ttype3 = ':'; 
		return 0; 
	case TT_WORD:
		m_subjectP = m_svalP;
		return 1;
	case '(':
		m_relations |= 2;
		if (nextToken() == TT_WORD) {
			m_subjectP = m_svalP;
			if (nextToken() == ')') {
				return 1;
		}	}
	default:
		fprintf(stderr, "*** Error (%d): Expecting word subject token. Found '%s'\n", m_lineno, m_svalP);
		error_exit(1);
	}
	return 1;
}

int
CLandscapeTokenStream::nextFactTriple(void) 
{
	m_startLineno = lineno();

	int ttype = nextToken();

	switch (ttype) {
	case TT_EOF:
		// End of section 
		return 0;
	case TT_WORD: 
		m_verbP = m_svalP;

		if (nextToken() != TT_WORD) {
			fprintf(stderr, "*** Error (%d): Expecting word token. Found '%s'\n", m_lineno, m_svalP);
			break;
		}

		m_objectP = m_svalP;

		ttype = nextToken();

		if (ttype == ':') {
			m_token1P = m_verbP;
			m_token2P = m_objectP; 
			m_ttype3  = ttype; 
			return 0; 
		} 

		if (ttype != TT_WORD) {
			fprintf(stderr, "*** Error (%d): Error in tuple parse. Found '%s'\n", m_lineno, m_svalP);
			break;
		}

		m_subjectP = m_svalP;

		return 1;
	default:	
		fprintf(stderr, "*** Error (%d): Expecting word token. Found '%s'\n", m_lineno, m_svalP);
	}
	error_exit(1);
	return 0;
}

void 
CLandscapeTokenStream::processSchemeAttributes(void) 
{

	CTAInput		*taP;
	int				ttype;
	CLandscapeObject	*targetP = 0;
	const char		*token1P;
	const char		*token2P;
	const char		*msgP;

	// 
	//  * id "{" {attribute} "}" *
	// 

	taP = m_taInputP;
	for (;;) {
	
		m_startLineno = lineno();

		ttype = nextToken();
		msgP  = 0;

		switch (ttype) {
		case TT_EOF:
			// End of section
			return; 
		case '(':
			// Relation or relation class 
			ttype = nextToken();
			if (ttype != TT_WORD) {
				msgP = "Expected ( <relation class name>";
				break;
			} 
			token1P = m_svalP;
			ttype   = nextToken();
			if (ttype != ')') {
				msgP = "Expected ( ... <)>";
				break;
			}
			ttype = nextToken();
			if (ttype != '{') {
				msgP = "Expected (...) <{>";
				break;
			}

			targetP = (CLandscapeObject *) taP->addRelationClass(token1P);
			break;
		case TT_WORD:
			// Entity class or entity 
			token1P = m_svalP;
			ttype   = nextToken();
			if (ttype != '{') {
				if (ttype != TT_WORD) {
					msgP = "Expecting ... <{>";
					break;
				}
				token2P = m_svalP;
				ttype   = nextToken();
				if (ttype != ':') {
					msgP = "Expecting section header or id <{>";
					break;
				}
				m_token1P = token1P; 
				m_token2P = token2P;
				m_ttype3  = ttype;
				// End of section 
				return;
			}
			targetP = (CLandscapeObject *) taP->addEntityClass(token1P);
			break;
		default:
			msgP = "Expecting object id";
		}
		if (msgP) {
			fprintf(stderr, "*** Error (%d):%s\n", lineno(), msgP);
			skipRecord();
			continue;
		} 
		parseAttributes(targetP); 
}	}

void 
CLandscapeTokenStream::processFactAttributes(void) 
{
	CTAInput			*taP;
	int					ttype;
	CLandscapeObject	*targetP = 0;
	const char			*token1P;
	const char			*token2P;
	const char			*token3P;
	const char			*msgP;
	CRelationClass		*rcP;
	CEntityInstance		*srcP, *dstP;

	// 
	//  * id "{" {attribute} "}" *
	// 

	taP = m_taInputP;


	for (;;) {

		m_startLineno = lineno();

		ttype = nextToken();
		msgP  = 0;

		switch (ttype) {
		case TT_EOF:
			// End of section
			return; 
		case '(':
		{
			// Relation or relation class 
			ttype = nextToken();
			if (ttype != TT_WORD) {
				msgP = "Expected ( <class> src dst )";
				break;
			} 
			token1P = m_svalP;
			ttype   = nextToken();
			if (ttype != TT_WORD) {
				msgP = "Expected ( token1 <src> dst)";
				break;
			}
			token2P = m_svalP;
			ttype   = nextToken();
			if (ttype != TT_WORD) {
				msgP = "Expected ( token1 token2 <dst>)";
				break;
			}
			token3P = m_svalP;
			ttype   = nextToken();
			if (ttype != ')') {
				msgP = "Expected (token1 token2 token3 <)>";
				break;
			}
			ttype = nextToken();
			if (ttype != '{') {
				msgP = "Expected (...) <{>";
				break;
			}
			
			rcP  = taP->getRelationClass(token1P);
			if (!rcP) {
				rcP  = taP->addRelationClass(token1P);
			}

			srcP = m_entityCacheP->get(token2P);
			if (!srcP) {
				// Create it
				srcP = taP->newCachedEntity(taP->getEntityBaseClass(), token2P);
			}

			dstP = m_entityCacheP->get(token3P);
			if (!dstP) {
				// Create it
				dstP = taP->newCachedEntity(taP->getEntityBaseClass(), token3P);
			}

			targetP = (CLandscapeObject *) srcP->getRelationTo(rcP, dstP);
			if (!targetP) {
				targetP = new CRelationInstance(rcP, srcP, dstP);
			}
			break;
		}
		case TT_WORD:
			// Entity class or entity 
			token1P = m_svalP;
			ttype   = nextToken();
			if (ttype != '{') {
				if (ttype != TT_WORD) {
					msgP = "Expecting token1 <{>";
					break;
				}
				token2P = m_svalP;
				ttype   = nextToken();
				if (ttype != ':') {
					msgP = "Expecting section header or id <{>";
					break;
				}
				m_token1P = token1P; 
				m_token2P = token2P;
				m_ttype3  = ttype;
				// End of section 
				return;
			}

			if (!stricmp(token1P, "$ROOT")) {
				targetP = taP->getRootInstance();
			} else {
				targetP = m_entityCacheP->get(token1P);
				if (!targetP) {
					// Create it
					targetP = taP->newCachedEntity(taP->getEntityBaseClass(), token1P);
				}
			}
			break;
		default:
			msgP = "Expecting object id";
		}
		if (msgP) {
			fprintf(stderr, "*** Error (%d):%s\n", lineno(), msgP);
			skipRecord();
			continue;
		} 
		parseAttributes(targetP); 
}	} 


